import inspect
from typing import Set, List, Type, Callable


class Task:

    @property
    def name(self):
        return self.__class__.name

    @property
    def deps(self):
        return set()

    def run(self):
        raise NotImplementedError


def add_task(name=None, deps=None, exec_fun=None) -> Type:
    cls = type(name, (Task,), {"deps": deps, "name": name, "run": exec_fun})
    return cls


def task(name: str, deps: List | Set | None = None) -> Callable[..., Callable[[Type | Callable], Type]]:
    """
       Decorator for creating Task

       Parameters:
       - name (str): The name of the task.
       - deps (List | Set | None, optional): Set of dependencies for the task. If not provided, it defaults to an empty set.

       Returns:
       - Callable[..., Callable[[Type | Callable], Type]]: A decorator that takes either a class or a callable function and returns a new class or adds the function as a task.

       Usage:
       - To decorate a class as a task:
         ```python
         @task(name="my_task", deps=["dependency1", "dependency2"])
         class MyTask(Task):
             # Class body
         ```

       - To decorate a function as a task:
         ```python
         @task(name="my_function_task", deps={"dependency1", "dependency2"})
         def my_function():
             # Function body
         ```
       """
    def wrap(f: Type | Callable) -> Type:
        if inspect.isclass(f):
            if issubclass(f, Task):
                return f
            else:
                base_classes = [Task] + list(cls for cls in f.__bases__)
                attr = f.__dict__
                attr.deps = deps or attr.get('deps', set())
                return type(name or f.__name__, tuple(base_classes), dict(attr))
        elif inspect.isfunction(f):
            return add_task(name=f.__name__, deps=deps, exec_fun=f)
        else:
            raise RuntimeError("only support function or class")
    return wrap



import warnings
from unittest import TestCase

from parade_manage.common.node import Node

from parade_manage.common.dag import DAG

from parade_manage import ParadeManage
from parade_manage.utils import walk_modules, tree, show_check_info


class Test(TestCase):

    def setUp(self) -> None:
        warnings.filterwarnings("ignore", category=DeprecationWarning)

        tasks = {
            "g": [], "h": [], "d": ["g"],
            "e": ["g", "h"], "a": ["d", "e"],
            "b": ["e", "f"], "c": ["f"],
            "k": [], "f": []
        }
        self.tasks = {Node(k, k): [Node(v, v) for v in vs] for k, vs in tasks.items()}
        self.dag = DAG.from_graph(self.tasks)

    def test_circular_dag(self):
        tasks = {
            "g": ["a"], "a": ["g"]
        }
        nodes = {Node(k, k): [Node(v, v) for v in vs] for k, vs in tasks.items()}
        DAG.from_graph(nodes)

    def test_check(self):
        tasks = {
            "g": ["a"], "a": ["g"], "b": ["b", "h"],
            "c": [], "k": [], "d": ["k", "c", "k"],
            "k1": ["k3"], "k2": ["k1"], "k3": ["k2"]
        }
        show_check_info(tasks)

    def test_check_normal(self):
        tasks = {
            "a": [], "b": [], "c": [],
            "k": ["a", "b"]
        }
        show_check_info(tasks)

    def test_walk_modules(self):
        print(walk_modules("../parade_manage/common"))

    def test_dump(self):
        pass

    def test_tree(self):

        tasks = {"flow-1": ["a", "b", "c"], "a": []}

        tree(tasks, "flow-1")

    def test_leaf_nodes(self):
        self.assertCountEqual([n.name for n in self.dag.leaf_nodes], ["g", "h", "f", "k"])

    def test_root_nodes(self):
        self.assertCountEqual([n.name for n in self.dag.root_nodes], ["a", "b", "c", "k"])

    def test_isolated_nodes(self):
        self.assertCountEqual([n.name for n in self.dag.isolated_nodes], ["k"])

    def test_show_tree(self):
        m = ParadeManage("/path/to/project")
        m.tree(flow_name="test-tree")

    def test_show_table(self):
        m = ParadeManage("/path/to/project")
        m.show()

    def test_sort(self):
        a = Node("A", "A")
        b = Node("B", "B", 2)
        c = Node("C", "C", 6)
        d = Node("D", "D", 3)
        e = Node("E", "E")
        f = Node("F", "F", 8)

        dag = DAG()
        dag.add_node(a)
        dag.add_node(b)
        dag.add_node(c)
        dag.add_node(d)
        dag.add_node(e)
        dag.add_node(f)

        dag.add_edge(c, a)
        dag.add_edge(c, b)
        dag.add_edge(e, c)
        dag.add_edge(e, d)
        dag.add_edge(f, a)
        dag.add_edge(f, b)

        sorted_nodes = dag.list()
        self.assertCountEqual([n.name for n in sorted_nodes], ["D", "B", "A", "F", "C", "E"])



from __future__ import annotations

import heapq
from copy import deepcopy
from typing import Dict, Set, List, Callable
from collections import defaultdict

from parade_manage.common.node import Node, NodeId


class DAG:
    def __init__(self):
        self._nodes: Dict[NodeId, Node] = dict()
        self._graph: Dict[Node, Set[Node]] = defaultdict(set)
        self._reversed_graph: Dict[Node, Set[Node]] = defaultdict(set)

        self.in_degree: Dict[NodeId, int] = defaultdict(int)

    @property
    def node_map(self) -> Dict[NodeId, Node]:
        return self._nodes

    @property
    def nodes(self) -> List[Node]:
        return list(self._nodes.values())

    @property
    def graph(self) -> Dict[Node, Set[Node]]:
        return self._graph

    @property
    def reversed_graph(self) -> Dict[Node, Set[Node]]:
        return self._reversed_graph

    def get_node(self, node_id: NodeId) -> Node:
        return self._nodes[node_id]

    def add_node(self, node: Node):
        node_id = node.node_id

        if node_id not in self._nodes:
            self._nodes[node_id] = node
            self._graph[node] = set()
            self._reversed_graph[node] = set()
            self.in_degree[node_id] = 0

    def remove_node(self, node: Node):
        node_id = node.node_id
        del self._nodes[node_id]

        del self._graph[node]
        for node_id_set in self._graph.values():
            if node_id in node_id_set:
                node_id_set.remove(node)

        del self._reversed_graph[node]
        for dep_node_id_set in self._reversed_graph.values():
            if node_id in dep_node_id_set:
                dep_node_id_set.remove(node)

    def contains_node(self, node: Node) -> bool:
        return node.node_id in self._nodes

    def add_edge(self, node: Node, dep_node: Node):
        nid, did = node.node_id, dep_node.node_id
        if nid not in self._nodes or did not in self._nodes:
            raise KeyError('node does not exist')

        self._graph[dep_node].add(node)
        self._reversed_graph[node].add(dep_node)

        self.in_degree[nid] += 1

    def remove_edge(self, node: Node, dep_node: Node):
        nid, did = node.node_id, dep_node.node_id
        if nid not in self._nodes or did not in self._nodes:
            raise KeyError('node does not exist')

        if nid in self._graph[dep_node]:
            self._graph[dep_node].remove(node)
            self._reversed_graph[node].remove(dep_node)

            self.in_degree[nid] -= 1

    def contains_edge(self, node: Node, dep_node: Node) -> bool:
        nid, did = node.node_id, dep_node.node_id
        if nid not in self._nodes or did not in self._nodes:
            return False

        return node in self._graph[dep_node]

    def find_no_dep_ids(self, reversed_graph: Dict[Node, Set[Node]] = None) -> List[NodeId]:

        reversed_graph = reversed_graph or self._reversed_graph

        no_dep_ids = []
        for node, dep_node_set in reversed_graph.items():
            if len(dep_node_set) == 0:
                no_dep_ids.append(node.node_id)

        return no_dep_ids

    def topological_sort(self, graph: Dict[Node, Set[Node]] = None, reversed_graph: Dict[Node, Set[Node]] = None
                         ) -> List[Node]:
        no_dep_ids: List[NodeId] = self.find_no_dep_ids()

        graph = graph or self._graph
        graph = deepcopy(graph)

        reversed_graph = reversed_graph or self._reversed_graph
        reversed_graph = deepcopy(reversed_graph)

        queue = list(no_dep_ids)

        traversed_ids = set()

        while len(queue) > 0:
            nid = queue.pop(0)
            traversed_ids.add(nid)

            dnode: Node = self._nodes[nid]

            child_nodes: Set[Node] = graph.pop(nid)
            for node in child_nodes:

                reversed_graph[node].remove(dnode)

                if len(reversed_graph[node]) == 0:
                    queue.append(node)

        if len(traversed_ids) != len(self._nodes):
            raise ValueError('Graph is not acyclic')

        return [self._nodes[nid] for nid in traversed_ids]

    def bfs(self, start_nodes: List[Node] = None):
        if start_nodes:
            start_nodes_ids = [node.node_id for node in start_nodes]
        else:
            start_nodes_ids = self.find_no_dep_ids()

        assert all(nid in self._nodes for nid in start_nodes_ids)

        visited = set(start_nodes_ids)

        queue = [self._nodes[nid] for nid in start_nodes_ids]

        while len(queue) > 0:
            cur_node = queue.pop(0)
            yield cur_node

            for node in self.children(cur_node):
                if node.node_id not in visited:
                    visited.add(node.node_id)
                    queue.append(node)

    def children(self, node: Node) -> List[Node]:
        return list(self._graph[node])

    def predecessor(self, node: Node) -> List[Node]:
        return list(self.reversed_graph[node])

    def successor(self, node: Node) -> List[Node]:
        return list(self.graph[node])

    def _traverse(self, nodes: List[Node], apply: Callable) -> Dict[Node, List[Node]]:
        all_deps = {}

        queue = nodes

        while len(queue) > 0:
            node = queue.pop()

            predecessor_nodes = apply(node)

            all_deps[node] = predecessor_nodes
            queue.extend(predecessor_nodes)

        return all_deps

    def all_predecessor(self, nodes: List[Node]) -> Dict[Node, List[Node]]:
        """
        get all predecessor node
        """
        return self._traverse(nodes, apply=self.predecessor)

    def all_successor(self, nodes: List[Node]) -> Dict[Node, List[Node]]:
        """
        get all successor node
        """
        return self._traverse(nodes, apply=self.successor)

    @property
    def isolated_nodes(self) -> List[Node]:
        """
        no predecessor and no successor
        :return: node
        """
        no_pred_nodes = [node for node, pnodes in self.reversed_graph.items() if len(pnodes) == 0]
        no_suc_nodes = [node for node, snodes in self.graph.items() if len(snodes) == 0]

        isolated_node = set(no_pred_nodes) & set(no_suc_nodes)

        return list(isolated_node)

    @property
    def root_nodes(self) -> List[Node]:
        no_pred_nodes = {node for node, pnodes in self.reversed_graph.items() if len(pnodes) == 0}
        return list(no_pred_nodes)

    @property
    def leaf_nodes(self) -> List[Node]:
        no_suc_nodes = {node for node, snodes in self.graph.items() if len(snodes) == 0}
        return list(no_suc_nodes)

    @classmethod
    def from_reversed_graph(cls, reversed_graph: Dict[Node, Set[Node] | List[Node]]) -> DAG:
        dag = DAG()
        for node, deps_nodes in reversed_graph.items():
            dag.add_node(node)
            for deps_node in deps_nodes:
                dag.add_node(deps_node)
                dag.add_edge(node, deps_node)

        return dag

    @classmethod
    def from_graph(cls, graph: Dict[Node, Set[Node] | List[Node]]) -> DAG:
        dag = DAG()
        for node, suc_nodes in graph.items():
            dag.add_node(node)
            for suc_node in suc_nodes:
                dag.add_node(suc_node)
                dag.add_edge(suc_node, node)

        return dag

    def list(self):
        return self._sort(self)
    @staticmethod
    def _sort(dag: DAG) -> List[Node]:
        queue = []
        for node_id, degree in dag.in_degree.items():
            if degree == 0:
                node = dag.get_node(node_id)
                heapq.heappush(queue, (node.priority * -1, node))  # max heap

        sorted_nodes = []
        in_degree = dag.in_degree.copy()

        while queue:
            priority, node = heapq.heappop(queue)
            sorted_nodes.append(node)

            succ_nodes = dag.successor(node)
            for node in succ_nodes:
                node_id = node.node_id
                in_degree[node_id] -= 1
                if in_degree[node_id] == 0:
                    heapq.heappush(queue, (node.priority * -1, node))

        if len(sorted_nodes) != len(dag.graph):
            raise ValueError("Graph contains a cycle")

        return sorted_nodes




from typing import Any, Dict

from parade_manage.common.dag import DAG


class Plugin:

    @property
    def plugin_name(self) -> str:
        raise NotImplementedError

    def run(self, dag: DAG) -> Any:
        raise NotImplementedError


class PluginRegistry:
    PLUGINS: Dict[str, Plugin] = dict()

    @staticmethod
    def register(plugin: Plugin):
        PluginRegistry.PLUGINS[plugin.plugin_name] = plugin

    @staticmethod
    def unregister(plugin_name: str):
        PluginRegistry.PLUGINS.pop(plugin_name, None)

    @staticmethod
    def load():
        return PluginRegistry.PLUGINS



# -*- coding: utf-8 -*-

"""
parade manager for managing `parade`
"""
from __future__ import annotations

import os
import sys

import yaml
import prettytable as pt

from typing import List, Type, Dict
from datetime import datetime

from parade.core.task import Task as ParadeTask
from parade_manage.common.node import Node
from .constants import FLAG_NODE_PRIORITY
from .plugin import Plugin, PluginRegistry

from .utils import iter_classes, tree, show_check_info
from .common.dag import DAG


class ParadeManage:

    def __init__(self, project_path: str = None, env: str = None):
        self.env = env
        self.project_path = self.init_context(project_path, env)

        self.dag: DAG = self.init_dag()
        self.plugins: Dict[str, Plugin] = {}

    def change_env(self, env: str):
        self.env = env
        self.init_context(self.project_path, env)
        print("change env to {}".format(env))

    @property
    def current_env(self) -> str:
        return os.environ.get("PARADE_PROFILE", "default")

    @property
    def task_map(self) -> Dict[str, Node]:
        """
        return task-name -> task
        """
        return {node.name: node for node in self.dag.nodes}

    @property
    def project(self) -> str:
        """
        :return: current project name
        """
        return self._get_project_name()

    def __repr__(self) -> str:
        return '<ParadeManager(project_path={}, env={})>'.format(self.project_path, self.env)

    def init_context(self, project_path: str = None, env: str = None) -> str:
        """
        init project context
        :param env: environment of project
        :param project_path: target project path
        :return: current project path
        """
        project_path = os.path.expanduser(project_path) if project_path is not None else os.getcwd()
        os.chdir(project_path)  # change project root path
        sys.path.insert(0, os.getcwd())

        if env:
            os.environ["PARADE_PROFILE"] = env

        return project_path

    def init_dag(self) -> DAG:

        project_name = self.project
        task_classes = iter_classes(ParadeTask, project_name + ".task")

        name_to_instance = self.init_task_classes(task_classes)

        reversed_graph: Dict[ParadeTask, List[ParadeTask]] = dict()

        for task_instance in name_to_instance.values():
            reversed_graph[task_instance] = []

            for deps_name in task_instance.deps:
                deps_task = name_to_instance[deps_name]
                reversed_graph[task_instance].append(deps_task)

        return self.to_dag(reversed_graph)

    def init_task_classes(self, task_classes: List[Type]) -> Dict[str, ParadeTask]:
        name_to_instance = {}
        for task_class in task_classes:
            task_instance = task_class()
            name_to_instance[task_instance.name] = task_instance

        return name_to_instance

    def _get_project_name(self) -> str:
        with open("parade.bootstrap.yml", "r") as f:
            conf = yaml.load(f, Loader=yaml.FullLoader)

        return conf['config']['name']

    @classmethod
    def to_dag(cls, reversed_graph: Dict[ParadeTask, List[ParadeTask]]) -> DAG:
        node_reversed_graph = {Node(k.name, k, cls._get_priority(k)): [Node(v.name, v, cls._get_priority(v)) for v in vs] for k, vs in reversed_graph.items()}
        dag = DAG.from_reversed_graph(node_reversed_graph)
        return dag
    @staticmethod
    def _get_priority(cls: ParadeTask) -> int:
        return getattr(cls, FLAG_NODE_PRIORITY, 0)

    def dump(self, target_tasks: str | List[str] = None, flow_name: str = None):
        """
        dump and generate file
        :param target_tasks: target tasks or None
        :param flow_name: flow name or None
        """
        flow_name = flow_name or "flow-" + datetime.now().strftime("%Y%m%d")

        if target_tasks is None:
            nodes = self.dag.nodes
        else:
            if isinstance(target_tasks, str):
                target_tasks = [target_tasks]

            current_nodes = [self.task_map[task] for task in target_tasks]
            nodes = self.dag.all_predecessor(current_nodes)

        tasks = [node.value for node in nodes]
        task_names = [task.name for task in tasks]
        deps = ["{task_name}->{task_deps}".format(task_name=task.name, task_deps=",".join(task.deps))
                for task in tasks if len(task.deps) > 0]

        data = {"tasks": task_names, "deps": deps}

        class IndentDumper(yaml.Dumper):
            def increase_indent(self, flow=False, indentless=False):
                return super(IndentDumper, self).increase_indent(flow, False)

        with open("./flows/" + flow_name + ".yml", "w") as f:
            yaml.dump(data, f, Dumper=IndentDumper, default_flow_style=False)

    def dump_with_prefix(self, prefix: str, flow_name: str = None):
        target_tasks = [task_name for task_name in self.task_map.keys() if task_name.startswith(prefix)]

        assert len(target_tasks) > 0, f"does not find task with prefix `{prefix}`"
        self.dump(target_tasks, flow_name)

    def dump_like(self, names: List[str], flow_name: str = None):
        target_tasks = set()
        for task_name in self.task_map.keys():
            for name in names:
                if task_name in name:
                    target_tasks.add(task_name)

        assert len(target_tasks) > 0, f"does not find task with names `{names}`"
        self.dump(list(target_tasks), flow_name)

    def tree(self, flow_name: str, task_names: List = None):
        """
        show task
        :param flow_name: name of flow
        :param task_names: task names or None
        """
        if task_names is None or len(task_names) == 0:
            nodes = self.dag.nodes
        else:
            nodes = self.dag.all_successor([self.task_map[task_name] for task_name in task_names])

        task_map: Dict[str, List[str]] = dict()
        for node in nodes:
            task: ParadeTask = node.value
            children = list(task.deps)
            task_map[task.name] = children

        task_map[flow_name] = list(task_map.keys())

        tree(task_map, flow_name)

    def show(self, task_names: List[str] = None, keyword: str = None):
        """show task in table"""
        if task_names is None or len(task_names) == 0:
            nodes = self.dag.nodes
        else:
            nodes = self.dag.all_successor([self.task_map[task_name] for task_name in task_names])

        tb = pt.PrettyTable()

        tb.field_names = ["name", "deps", "description"]

        for node in nodes:
            task = node.value
            description = (getattr(task, "description", "") or getattr(task, "describe", "") or
                           getattr(task, "__doc__", "")) or ""
            if keyword:
                if self._filter_item([task.name, description], keyword):
                    tb.add_row([task.name, "\n".join(task.deps), description.strip()], divider=True)
            else:
                tb.add_row([task.name, "\n".join(task.deps), description.strip()], divider=True)

        print(f"Total: {len(tb.rows)}")
        print(tb)

    def _filter_item(self, items: List[str], keyword: str) -> bool:
        """filter item"""
        return len([item for item in items if keyword is not None and keyword in item]) > 0

    @property
    def isolated_tasks(self) -> List[str]:
        """
        no predecessor and no successor
        :return: task name
        """
        return [node.name for node in self.dag.isolated_nodes]

    @property
    def leaf_tasks(self) -> List[str]:
        """
        :return: task name
        """
        return [node.name for node in self.dag.leaf_nodes]

    @property
    def root_tasks(self) -> List[str]:
        """
        :return: task name
        """
        return [node.name for node in self.dag.root_nodes]

    def check(self):
        reversed_graph = self.dag.reversed_graph
        tasks = {t.name: [n.name for n in d] for t, d in reversed_graph.items()}

        show_check_info(tasks)

    def build(self, target_tasks: str | List[str] = None):
        """generate flow"""
        return self.dump(target_tasks)

    def reload_plugins(self):
        self.plugins = PluginRegistry.load()

    def execute(self, cmd):
        plugin: Plugin = self.plugins.get(cmd)
        if plugin:
            raise KeyError(f"plugin {cmd} not found")
        return plugin.run(self.dag)



# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='parade-manage',
    version='0.0.2.11',
    author='Pan Chen',
    author_email='chenpan9012@gmail.com',
    description='A manage module of parade',
    url='https://github.com/cpzt/parade-manage',
    install_requires=['parade'],
    zip_safe=False,
    python_requires='>=3.4',
    include_package_data=True,
    platforms=['any'],
    packages=find_packages(),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX'
    ]
)



from __future__ import annotations
from typing import Any, List, Collection

NodeId = int


class Node:

    def __init__(self, name: str, value: Any, priority: int = 0):
        self.name = name
        self.value = value
        # The priority of the node.
        # Higher values indicate higher priority (default is 0).
        self.priority = self._validate_priority(priority)

    def set_priority(self, priority: int) -> None:
        assert priority >= 0, "`priority` must be a non-negative integer"
        self.priority = self._validate_priority(priority)

    @staticmethod
    def _validate_priority(value: int) -> int:
        if not isinstance(value, int) or value < 0:
            raise ValueError("`priority` must be a non-negative integer")
        return value

    @property
    def node_id(self) -> NodeId:
        return NodeId(hash(self.name))

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return f"Node(name={self.name}, value={self.value}, priority={self.priority})"

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.priority > other.priority

    @staticmethod
    def build_list(elements: Collection) -> List[Node]:
        if len(elements) == 0:
            return []

        if isinstance(elements, (list, set)):
            nodes = []

            for element in elements:
                if isinstance(elements[0], (list, tuple, set)) and len(elements[0]) >= 2:
                    node = Node(element[0], element[1])
                elif isinstance(elements[0], (list, tuple, set)) and len(elements[0]) == 1:
                    node = Node(element[0], element[0])
                else:
                    node = Node(element, element)
                nodes.append(node)
            return nodes
        else:
            raise ValueError(f"only support list or set. elements is {elements}.")



from __future__ import annotations

try:
    # Python <= 3.9
    from collections import Iterable
except ImportError:
    # Python > 3.9
    from collections.abc import Iterable

from collections import defaultdict

import inspect
from importlib import import_module

from pkgutil import iter_modules
from typing import List, Type, Callable, Dict, Set, Tuple, Union


def walk_modules(path: str) -> List:
    """
    :param path: e.g.  your.project
    :return : e.g. ["your.project.some"]
    """
    mods = []

    mod = import_module(path)
    mods.append(mod)
    if not hasattr(mod, '__path__'):
        return []

    for finder, sub_path, is_pkg in iter_modules(mod.__path__):
        full_path = path + "." + sub_path
        if is_pkg:
            mods.extend(walk_modules(full_path))
        else:
            mod = import_module(full_path)
            mods.append(mod)

    return mods


def iter_classes(base_class: Type, *modules, class_filter: Callable = None):
    for root_module in modules:
        try:
            mods = walk_modules(root_module)
        except Exception as e:
            raise e

        for mod in mods:
            for class_obj in vars(mod).values():
                if inspect.isclass(class_obj) and issubclass(class_obj, base_class) \
                        and class_obj.__module__ == mod.__name__:
                    if not class_filter or class_filter(class_obj):
                        yield class_obj


def flatten(items, ignore_types=(bytes, str), ignore_flags=('', None)):
    for item in items:
        if item in ignore_flags:
            continue
        if isinstance(item, Iterable) and not isinstance(item, ignore_types):
            yield from flatten(item)
        else:
            yield item


def tree(item_map: Dict[str, List[str]], name: str, prefix: str = "", is_root: bool = True, is_tail: bool = True):
    if is_root:
        print(prefix + " " + name)
    else:
        print(prefix + ("└── " if is_tail else "├── ") + name)

    children = item_map.get(name, [])

    if len(children) > 0:
        last_child = children[-1]
        rest_child = children[0:-1]
        for child in rest_child:
            tree(item_map, child, prefix + ("    " if is_tail else "│   "), False, False)
        tree(item_map, last_child, prefix + ("    " if is_tail else "│   "), False, True)


def find_cycles(graph: Dict[str, List[str] | Set[str]]) -> List[List[str]]:
    visited = set()
    recursion_stack = set()
    cycles = []

    def dfs(node, path):
        if node in recursion_stack:
            cycles.append(path + [node])
            return

        if node in visited:
            return

        visited.add(node)
        recursion_stack.add(node)

        if node in graph:
            for neighbor in graph[node]:
                dfs(neighbor, path + [node])

        recursion_stack.remove(node)

    for node in graph:
        dfs(node, [])

    return cycles


def check(tasks: Dict[str, List[str] | Set[str]]) -> Tuple[Dict, Dict, Dict, List]:

    non_deps_tasks: Dict[str, Set[str]] = defaultdict(set)
    duplicate_tasks: Dict[str, Set[Tuple[str, int]]] = defaultdict(set)
    circular_tasks: List[List[str]] = find_cycles(tasks)

    for task, deps in tasks.items():
        for dp in deps:
            # check for invalid dependencies
            if dp not in tasks:
                non_deps_tasks[task].add(dp)

            # check for duplicate dependencies
            if deps.count(dp) > 1:
                duplicate_tasks[task].add((dp, deps.count(dp)))

    non_deps_tasks = {k: set(v) for k, v in non_deps_tasks.items()}
    duplicate_tasks = {k: set(v) for k, v in duplicate_tasks.items()}

    return tasks, non_deps_tasks, duplicate_tasks, circular_tasks


def show_check_info(tasks: Dict[str, List[str] | Set[str]]):
    deps, non_deps, duplicate, circular = check(tasks)

    if len(non_deps) == 0 and len(duplicate) == 0 and len(circular) == 0:
        print(f"Total: {len(deps)}, PASS")
        return
    else:
        print(f"Total: {len(deps)}")

    dividing_line = "---*---" * 8
    print(dividing_line)

    if len(non_deps) > 0:
        print("[Invalid Dependencies]")
        for k, v in non_deps.items():
            print(k, ' ==>  ', v)
        print("\n" + dividing_line)

    if len(duplicate) > 0:
        print('[Duplicate Dependencies]')
        for k, v in duplicate.items():
            print(k, ' ==>  ', v)
        print("\n" + dividing_line)

    if len(circular) > 0:
        print('[Circular Dependencies]')
        for cycle in circular:
            print(" -> ".join(cycle))
        print("\n" + dividing_line)


def add_tag(cls, tag: Union[str, Iterable[str]]):
    if tag:
        if not hasattr(cls, "__tags__"):
            cls.__tags__ = set()
        if isinstance(tag, str):
            tags = [tag]
        else:
            tags = tag
        for tag in tags:
            cls.__tags__.add(tag)

def tag(tags: Iterable[str]):
    def wrapper(cls):
        add_tag(cls, tags)
        return cls

    return wrapper
